"""
====================================
Working with IRIS spectrograph files
====================================

In this example, we will showcase how to use ``irispy-lmsal`` to open, crop and plot IRIS spectrograph data.
"""

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pooch
from astropy.coordinates import SkyCoord, SpectralCoord
from sunpy.coordinates.frames import Helioprojective

from irispy.io import read_files

###############################################################################
# We start with getting the data.
# This is done by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` to keep this example self contained
# but using your browser will also work.

raster_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2018/01/02/20180102_153155_3610108077/iris_l2_20180102_153155_3610108077_raster.tar.gz",
    known_hash="0ec2b7b20757c52b02e0d92c27a5852b6e28759512c3d455f8b6505d4e1f5cd6",
)

###############################################################################
# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float32, the data values are no longer in DN,
# but in scaled integer units that start at -2$^{16}$/2.

raster = read_files(raster_filename, memmap=True, uncertainty=False)
# Provide an overview of the data
print(raster)
# Will give us all the keys that corresponds to wavelengths.
print(raster.keys())

###############################################################################
# We will now access one wavelength window and plot it.

mg_ii = raster["Mg II k 2796"][0]

plt.figure()
mg_ii.plot()

plt.show()

###############################################################################
# We can index to get the first index and plot it.

plt.figure()
mg_ii[0].plot()

plt.show()

###############################################################################
# Or we can get the spectrum directly.

plt.figure()
mg_ii[120, 200].plot()

plt.show()

###############################################################################
# The default plots take the units and labels from the FITS WCS information,
# and often do not come in the most useful units (e.g. wavelengths in metres).
# We can read the wavelengths of the Mg window by calling axis_world_coords for wl
# (wavelength), and redo the plot with a better scale:

(mg_wave,) = mg_ii.axis_world_coords("wl")
mg_wave.to("m")

fig, ax = plt.subplots()
# Here we index the data directly instead of other methods to get access to the data and plot it "raw"
plt.plot(mg_wave.to("nm"), mg_ii.data[120, 200])

plt.show()

###############################################################################
#  When we use the underlying data directly, we lose all the metadata and WCS information,
# so we would suggest not doing it normally but there will be times you will need to do this.
# If you are unfamiliar with WCS, the following links are quite useful:
# The base comes from astropy: https://docs.astropy.org/en/stable/wcs/index.html
# The plotting makes use of WCSAxes: https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html
# Some of the higher-level utils are via NDCube, e.g. coordinate transformations: https://docs.sunpy.org/projects/ndcube/en/v2.0.0rc2/coordinates.html.
# We can also improve on the default spectrogram plot by adjusting some options:

plt.figure()
mg_ii[0].plot(aspect="auto", cmap="irissjiNUV")
ax = plt.gca()
ax.coords[0].set_major_formatter("x.x")
ax.coords[0].set_format_unit(u.nm)
ax.set_xlabel("Wavelength (nm)")
ax.set_ylabel("Helioprojective latitude [Solar Y] (arcsec)")
# Remove longitude ticks
ax.coords[2].set_ticks([] * u.degree)

plt.show()

###############################################################################
# What is the wavelength position that corresponds to Mg II k core (279.63 nm)?

wcs_loc = mg_ii.wcs.world_to_pixel(
    SpectralCoord(279.63, unit=u.nm),
    SkyCoord(0 * u.arcsec, 0 * u.arcsec, frame=Helioprojective),
)
mg_index = int(np.round(wcs_loc[0]))
print(mg_index)

###############################################################################
# Now we will plot spectroheliogram for Mg II k core wavelength.
# We can use a new feature in ndcube 2.0 called crop to get this information
# that will require a SkyCoord and SpectralCoord object from astropy.
# In this case, we only need to use a SpectralCoord.

lower_corner = [SpectralCoord(280, unit=u.nm), None]
upper_corner = [SpectralCoord(280, unit=u.nm), None]
mg_crop = mg_ii.crop(lower_corner, upper_corner)

ax = mg_crop.plot()
plt.xlabel("Solar X")
plt.ylabel("Solar Y")

plt.show()

###############################################################################
# Imagine there's a really cool feature at (-338", 275"), how can you plot
# the spectrum at that location?

lower_corner = [None, SkyCoord(-338 * u.arcsec, 275 * u.arcsec, frame=Helioprojective)]
upper_corner = [None, SkyCoord(-338 * u.arcsec, 275 * u.arcsec, frame=Helioprojective)]
mg_ii.crop(lower_corner, upper_corner).plot()

plt.show()

###############################################################################
#  Now, you may also be interested in knowing the time that was this observation taken.
# There is some information in ``.meta``.

print(mg_ii.meta)

###############################################################################
# But this is mostly about the observation in general.
# Times of individual scans are saved in .extra_coords['time'].
# Getting access to it can be done in the following  way:

# The first index is to escape the tuple that ``axis_world_coords`` returns
print(mg_ii.axis_world_coords("time", wcs=mg_ii.extra_coords)[0][0].isot)



import astropy.units as u
import numpy as np

AXIS = [
    (
        "custom:pos.helioprojective.lon",
        "custom:pos.helioprojective.lat",
        "time",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
        "custom:CUSTOM",
    ),
    ("custom:pos.helioprojective.lon", "custom:pos.helioprojective.lat"),
    ("custom:pos.helioprojective.lon", "custom:pos.helioprojective.lat"),
]


def test_world_axis_physical_types_SJICube_2832(SJICube_2832):
    assert np.all(SJICube_2832.dimensions == [10, 40, 37] * u.pix)
    assert SJICube_2832.array_axis_physical_types == AXIS


def test_world_axis_physical_types_SJICube_2796(SJICube_2796):
    assert np.all(SJICube_2796.dimensions == [62, 40, 37] * u.pix)
    assert SJICube_2796.array_axis_physical_types == AXIS


def test_world_axis_physical_types_SJICube_1400(SJICube_1400):
    assert np.all(SJICube_1400.dimensions == [62, 40, 37] * u.pix)
    assert SJICube_1400.array_axis_physical_types == AXIS


def test_world_axis_physical_types_SJICube_1330(SJICube_1330):
    assert np.all(SJICube_1330.dimensions == [52, 40, 37] * u.pix)
    assert SJICube_1330.array_axis_physical_types == AXIS



import astropy.units as u
import sunpy.visualization.colormaps as cm  # NOQA: F401
from mpl_animators import ArrayAnimatorWCS
from ndcube.visualization.mpl_plotter import MatplotlibPlotter

__all__ = ["_set_axis_colors", "Plotter", "CustomArrayAnimatorWCS"]


def _set_axis_colors(ax):
    """
    Private method to set the axis colors and labels for IRIS SJI and Raster
    data.
    """
    if isinstance(ax, ArrayAnimatorWCS):
        ax = ax.axes
    for axis in ax.coords:
        if axis.default_label.lower() in ["wavelength", "wave", "em.wl"]:
            axis.set_format_unit(u.nm)
            axis.set_major_formatter("x.x")
            axis.set_axislabel("em.wl [$\\mathrm{nm}$]")
        elif axis.default_label.lower() in [
            "latitude",
            "lat",
            "custom:pos.helioprojective.lat",
        ]:
            _extracted_from__set_axis_colors_14(axis, "red")
        elif axis.default_label.lower() in [
            "longitude",
            "lon",
            "custom:pos.helioprojective.lon",
        ]:
            _extracted_from__set_axis_colors_14(axis, "black")


def _extracted_from__set_axis_colors_14(axis, color):
    axis.set_axislabel(axis.default_label, color=color)
    axis.set_ticklabel_position("all")
    axis.set_ticklabel(color)


class CustomArrayAnimatorWCS(ArrayAnimatorWCS):
    def update_plot_2d(self, val, im, slider):
        """
        Update the image plot.
        """
        self.axes.reset_wcs(wcs=self.wcs, slices=self.slices_wcsaxes)
        im.set_array(self.data_transposed)
        if self.clip_interval is not None:
            vmin, vmax = self._get_2d_plot_limits()
            im.set_clim(vmin, vmax)
        slider.cval = val
        _set_axis_colors(self.axes)


class Plotter(MatplotlibPlotter):
    def plot(self, *args, **kwargs):
        return super().plot(*args, **kwargs)

    def _animate_cube(
        self,
        wcs,
        plot_axes=None,
        axes_coordinates=None,
        axes_units=None,
        data_unit=None,
        **kwargs,
    ):
        # Derive inputs for animation object and instantiate.
        data, wcs, plot_axes, coord_params = self._prep_animate_args(wcs, plot_axes, axes_units, data_unit)
        ax = CustomArrayAnimatorWCS(data, wcs, plot_axes, coord_params=coord_params, **kwargs)
        # We need to modify the visible axes after the axes object has been created.
        # This call affects only the initial draw
        self._apply_axes_coordinates(ax.axes, axes_coordinates)
        # This changes the parameters for future iterations
        for hidden in self._not_visible_coords(ax.axes, axes_coordinates):
            param = ax.coord_params.get(hidden, {})
            param["ticks"] = False
            ax.coord_params[hidden] = param
        return ax



"""
====================
A quick and easy SJI
====================

In this example we will show how to plot a South Pole SJI dataset.

You can get IRIS data with co-aligned SDO data (and more) from https://iris.lmsal.com/search/
"""

import matplotlib.pyplot as plt
import pooch

from irispy.io import read_files

###############################################################################
# We start with getting the data.
# This is done by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` as to keep this example self contained
# but using your browser will also work.

sji_filename = pooch.retrieve(
    "https://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2023/02/11/20230211_083601_3880012095/iris_l2_20230211_083601_3880012095_SJI_2832_t000.fits.gz",
    known_hash="a4b4f108ed67aded876bb167f24f093e8401d836f7deb66c97a36da8dd226064",
)

###############################################################################
# We will now open the slit-jaw imager (SJI) file we just downloaded.

sji_2832 = read_files(sji_filename)
# Printing will give us an overview of the file.
print(sji_2832)

###############################################################################
# We will now plot the IRIS SJI data.
# You can also change the axis labels and ticks if you so desire.
# `WCSAxes provides us an API we can use. <https://docs.astropy.org/en/stable/visualization/wcsaxes/index.html>`__

# Note that the .get_animation() is used to animate this example and is not required normally.
ax = sji_2832.plot().get_animation()
plt.xlabel("Helioprojective Longitude (Solar-X) [arcsec]")
plt.ylabel("Helioprojective Latitude (Solar-Y) [arcsec]")
plt.title(f"IRIS SJI {sji_2832.meta['TWAVE1']}", pad=20)

plt.show()



import copy

import astropy.units as u
import numpy as np
from astropy.io import fits
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS

import irispy.data.test
from irispy import SpectrogramCube, SpectrogramCubeSequence, utils
from irispy.io.spectrograph import read_spectrograph_lvl2

testpath = irispy.data.test.rootdir
# Arrays of DN
SOURCE_DATA_DN = np.array(
    [
        [[0.563, 1.132, -1.343], [-0.719, 1.441, 1.566]],
        [[0.563, 1.132, -1.343], [-0.719, 1.441, 1.566]],
    ],
)
SOURCE_UNCERTAINTY_DN = np.sqrt(SOURCE_DATA_DN)
# Arrays relating SOURCE_DATA_DN to photons in NUV and FUV
SOURCE_DATA_PHOTONS_NUV = np.array(
    [
        [[10.134, 20.376, -24.174], [-12.942, 25.938, 28.188]],
        [[10.134, 20.376, -24.174], [-12.942, 25.938, 28.188]],
    ],
)
SOURCE_DATA_PHOTONS_FUV = np.array(
    [
        [[2.252, 4.528, -5.372], [-2.876, 5.764, 6.264]],
        [[2.252, 4.528, -5.372], [-2.876, 5.764, 6.264]],
    ],
)
SOURCE_UNCERTAINTY_PHOTONS_NUV = np.sqrt(SOURCE_DATA_PHOTONS_NUV)
SOURCE_UNCERTAINTY_PHOTONS_FUV = np.sqrt(SOURCE_DATA_PHOTONS_FUV)
time_dim_len = SOURCE_DATA_DN.shape[0]
single_exposure_time = 2.0
EXPOSURE_TIME = u.Quantity(np.zeros(time_dim_len) + single_exposure_time, unit=u.s)
# Define an sample wcs object
h0 = {
    "CTYPE1": "WAVE    ",
    "CUNIT1": "Angstrom",
    "CDELT1": 0.2,
    "CRPIX1": 0,
    "CRVAL1": 10,
    "NAXIS1": 3,
    "CTYPE2": "HPLT-TAN",
    "CUNIT2": "deg",
    "CDELT2": 0.5,
    "CRPIX2": 2,
    "CRVAL2": 0.5,
    "NAXIS2": 2,
    "CTYPE3": "HPLN-TAN",
    "CUNIT3": "deg",
    "CDELT3": 0.4,
    "CRPIX3": 2,
    "CRVAL3": 1,
    "NAXIS3": 2,
}
wcs0 = WCS(header=h0, naxis=3)
# Define sample meta
meta0 = {"detector type": "FUV", "OBSID": 1, "spectral window": "C II 1336"}
# Define sample extra coords
extra_coords0 = [
    ("time", 0, Time("2017-01-01") + TimeDelta(np.arange(time_dim_len), format="sec")),
    ("exposure time", 0, EXPOSURE_TIME),
]
extra_coords1 = [
    (
        "time",
        0,
        (Time("2017-01-01") + TimeDelta(np.arange(time_dim_len, time_dim_len * 2), format="sec")),
    ),
    ("exposure time", 0, EXPOSURE_TIME),
]
# Define SpectrogramCubes in various units.
spectrogram_DN0 = SpectrogramCube(
    SOURCE_DATA_DN,
    wcs0,
    SOURCE_UNCERTAINTY_DN,
    utils.DN_UNIT["FUV"],
    meta0,
)
spectrogram_DN0.extra_coords.add(*extra_coords0[0])
spectrogram_DN0.extra_coords.add(*extra_coords0[1])
spectrogram_photon0 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV,
    u.photon,
    meta0,
)
spectrogram_DN0.extra_coords.add(*extra_coords0[0])
spectrogram_DN0.extra_coords.add(*extra_coords0[1])
spectrogram_DN_per_s0 = SpectrogramCube(
    SOURCE_DATA_DN / single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_DN / single_exposure_time,
    utils.DN_UNIT["FUV"] / u.s,
    meta0,
)
spectrogram_DN_per_s0.extra_coords.add(*extra_coords0[0])
spectrogram_DN_per_s0.extra_coords.add(*extra_coords0[1])
spectrogram_photon_per_s0 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV / single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV / single_exposure_time,
    u.photon / u.s,
    meta0,
)
spectrogram_photon_per_s0.extra_coords.add(*extra_coords0[0])
spectrogram_photon_per_s0.extra_coords.add(*extra_coords0[1])
spectrogram_DN1 = SpectrogramCube(
    SOURCE_DATA_DN,
    wcs0,
    SOURCE_UNCERTAINTY_DN,
    utils.DN_UNIT["FUV"],
    meta0,
)
spectrogram_DN1.extra_coords.add(*extra_coords1[0])
spectrogram_DN1.extra_coords.add(*extra_coords1[1])
spectrogram_photon1 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV,
    u.photon,
    meta0,
)
spectrogram_photon1.extra_coords.add(*extra_coords1[0])
spectrogram_photon1.extra_coords.add(*extra_coords1[1])
spectrogram_DN_per_s1 = SpectrogramCube(
    SOURCE_DATA_DN / single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_DN / single_exposure_time,
    utils.DN_UNIT["FUV"] / u.s,
    meta0,
)
spectrogram_DN_per_s1.extra_coords.add(*extra_coords1[0])
spectrogram_DN_per_s1.extra_coords.add(*extra_coords1[1])
spectrogram_photon_per_s1 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV / single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV / single_exposure_time,
    u.photon / u.s,
    meta0,
)
spectrogram_DN0.extra_coords.add(*extra_coords0[0])
spectrogram_DN0.extra_coords.add(*extra_coords0[1])
spectrogram_photon_per_s_per_s0 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV / single_exposure_time / single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV / single_exposure_time / single_exposure_time,
    u.photon / u.s / u.s,
    meta0,
)
spectrogram_photon_per_s_per_s0.extra_coords.add(*extra_coords0[0])
spectrogram_photon_per_s_per_s0.extra_coords.add(*extra_coords0[1])
spectrogram_photon_s0 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV * single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV * single_exposure_time,
    u.photon * u.s,
    meta0,
)
spectrogram_photon_s0.extra_coords.add(*extra_coords0[0])
spectrogram_photon_s0.extra_coords.add(*extra_coords0[1])
spectrogram_photon_per_s_per_s1 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV / single_exposure_time / single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV / single_exposure_time / single_exposure_time,
    u.photon / u.s / u.s,
    meta0,
)
spectrogram_photon_per_s_per_s1.extra_coords.add(*extra_coords1[0])
spectrogram_photon_per_s_per_s1.extra_coords.add(*extra_coords1[1])
spectrogram_photon_s1 = SpectrogramCube(
    SOURCE_DATA_PHOTONS_FUV * single_exposure_time,
    wcs0,
    SOURCE_UNCERTAINTY_PHOTONS_FUV * single_exposure_time,
    u.photon * u.s,
    meta0,
)
spectrogram_photon_s1.extra_coords.add(*extra_coords1[0])
spectrogram_photon_s1.extra_coords.add(*extra_coords1[1])
# Define meta dict for an SpectrogramCubeSequence
meta_seq = {
    "detector type": "FUV",
    "spectral window": "C II 1336",
    "brightest wavelength": 100,
    "min wavelength": 90,
    "max wavelength": 110,
}
# Define SpectrogramCubeSequences
sequence_DN = SpectrogramCubeSequence([spectrogram_DN0, spectrogram_DN1], meta_seq)
sequence_photon = SpectrogramCubeSequence([spectrogram_photon0, spectrogram_photon1], meta_seq)
sequence_DN_per_s = SpectrogramCubeSequence([spectrogram_DN_per_s0, spectrogram_DN_per_s1], meta_seq)
sequence_photon_per_s = SpectrogramCubeSequence([spectrogram_photon_per_s0, spectrogram_photon_per_s1], meta_seq)
sequence_photon_per_s_per_s = SpectrogramCubeSequence(
    [spectrogram_photon_per_s_per_s0, spectrogram_photon_per_s1],
    meta_seq,
)
sequence_photon_s = SpectrogramCubeSequence([spectrogram_photon_s0, spectrogram_photon_s1], meta_seq)


def test_fits_data_comparison(raster_file):
    """
    Make sure the data is the same in pyfits and irispy.
    """
    iris_l2_test_raster = read_spectrograph_lvl2(raster_file)
    with fits.open(raster_file) as hdulist:
        spectral_window1 = hdulist[0].header["TDESC1"]
        spectral_window2 = hdulist[0].header["TDESC2"]
        spectral_window3 = hdulist[0].header["TDESC3"]
        data1 = copy.deepcopy(hdulist[1].data)
        data2 = copy.deepcopy(hdulist[2].data)
        data3 = copy.deepcopy(hdulist[3].data)
        np.testing.assert_array_almost_equal(iris_l2_test_raster[spectral_window1].data[0].data, data1)
        np.testing.assert_array_almost_equal(iris_l2_test_raster[spectral_window2].data[0].data, data2)
        np.testing.assert_array_almost_equal(iris_l2_test_raster[spectral_window3].data[0].data, data3)



import logging
import tarfile
from pathlib import Path

from astropy.io import fits

__all__ = ["fitsinfo", "read_files"]


def fitsinfo(filename):
    """
    Prints information about the extension of a raster or SJI level 2 data
    file.

    Parameters
    ----------
    filename : str
        Filename to load.
    """
    with fits.open(filename) as hdulist:
        hdulist.info()
        hdr = hdulist[0].header
        msg = f"Observation description: {hdr['OBS_DESC']}"
        logging.info(msg)
        modifier = ""
        for i in range(hdr["NWIN"]):
            msg = f"Extension No. {i+1} stores data and header of {hdr[f'TDESC{i+1}']}: "
            logging.info(msg)
            if "SJI" not in hdr[f"TDET{i + 1}"]:
                modifier = f" ({hdr[f'TDET{i + 1}'][:3]})"
            msg = f"{hdr[f'TWMIN{i + 1}']:.2f} - {hdr[f'TWMAX{i + 1}']:.2f} AA{modifier}"
            logging.info(msg)


def read_files(filename, *, spectral_windows=None, uncertainty=False, memmap=True):
    """
    A wrapper function to read a raster or SJI level 2 data file.

    You can provide one SJI image or a one raster image or a list of raster images.

    If you mix raster and SJI images, the function will raise an error.

    Parameters
    ----------
    filename : `list of `str`, `str`, `pathlib.Path`
        Filename(s) to load.
        If given a string, will load that file.
        If given a list of strings, it will check they are all raster files and load them.
    spectral_windows: iterable of `str` or `str`
        Spectral windows to extract from files. Default=None, implies, extract all
        spectral windows.
    uncertainty : `bool`, optional
        If `True` (not the default), will compute the uncertainty for the data (slower and
        uses more memory). If `memmap=True`, the uncertainty is never computed.
    memmap : `bool`, optional
        If `True` (the default), will not load arrays into memory, and will only read from
        the file into memory when needed. This option is faster and uses a
        lot less memory. However, because FITS scaling is not done on-the-fly,
        the data units will be unscaled, not the usual data numbers (DN).

    Returns
    -------
    The corresponding `irispy.sji.SJICube` or `irispy.spectrogram.SpectrogramCube`.
    """
    from irispy.io.sji import read_sji_lvl2
    from irispy.io.spectrograph import read_spectrograph_lvl2

    if isinstance(filename, Path):
        filename = str(filename)
    if isinstance(filename, str | Path):
        if tarfile.is_tarfile(filename):
            path = Path(filename.replace(".tar.gz", ""))
            path.mkdir(parents=True, exist_ok=True)
            with tarfile.open(filename, "r") as tar:
                tar.extractall(path)
                filename = [path / file for file in tar.getnames()]
        else:
            filename = [filename]
    intrume = fits.getval(filename[0], "INSTRUME")
    all_instrume = [fits.getval(f, "INSTRUME") for f in filename]
    if any(intrume != i for i in all_instrume):
        msg = "You cannot mix raster and SJI files."
        raise ValueError(msg)
    if intrume == "SJI":
        if len(filename) > 1:
            msg = "You cannot load more than one SJI file at a time."
            raise ValueError(msg)
        return read_sji_lvl2(filename[0], memmap=memmap, uncertainty=uncertainty)
    if intrume == "SPEC":
        return read_spectrograph_lvl2(
            filename,
            spectral_windows=spectral_windows,
            memmap=memmap,
            uncertainty=uncertainty,
        )
    msg = f"Unsupported instrument: {intrume}"
    raise ValueError(msg)



"""
Tests that use fake data (fd) instead of the test files.
"""

import numpy as np
import pytest
from astropy import units as u
from astropy.time import Time
from astropy.wcs import WCS
from sunraster.extern.meta import Meta

from irispy import SJICube, utils

##############################################################################
# SJICube Fixtures
##############################################################################
data = np.array(
    [
        [[1, 2, 3, 4], [2, 4, 5, 3], [0, 1, 2, 3]],
        [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]],
    ],
)
data_2D = np.array([[1, 2, 3, 4], [2, 4, 5, 3]])
data_1D = np.array([1, 2])
data_4D = np.array(
    [
        [
            [[1, 2, 3, 4], [2, 4, 5, 3], [0, 1, 2, 3]],
            [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]],
        ],
        [
            [[1, 2, 3, 4], [2, 4, 5, 3], [0, 1, 2, 3]],
            [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]],
        ],
    ],
)
header = {
    "CTYPE1": "HPLN-TAN",
    "CUNIT1": "arcsec",
    "CDELT1": 0.4,
    "CRPIX1": 0,
    "CRVAL1": 0,
    "NAXIS1": 4,
    "CTYPE2": "HPLT-TAN",
    "CUNIT2": "arcsec",
    "CDELT2": 0.5,
    "CRPIX2": 0,
    "CRVAL2": 0,
    "NAXIS2": 3,
    "CTYPE3": "Time    ",
    "CUNIT3": "seconds",
    "CDELT3": 0.3,
    "CRPIX3": 0,
    "CRVAL3": 0,
    "NAXIS3": 2,
}
wcs = WCS(header=header, naxis=3)
header_2D = {
    "CTYPE1": "HPLN-TAN",
    "CUNIT1": "arcsec",
    "CDELT1": 0.4,
    "CRPIX1": 0,
    "CRVAL1": 0,
    "NAXIS1": 4,
    "CTYPE2": "HPLT-TAN",
    "CUNIT2": "arcsec",
    "CDELT2": 0.5,
    "CRPIX2": 0,
    "CRVAL2": 0,
    "NAXIS2": 3,
}
wcs_2D = WCS(header=header_2D, naxis=2)
header_4D = {
    "CTYPE1": "Time    ",
    "CUNIT1": "seconds",
    "CDELT1": 0.4,
    "CRPIX1": 0,
    "CRVAL1": 0,
    "NAXIS1": 4,
    "CTYPE2": "HPLT-TAN",
    "CUNIT2": "arcsec",
    "CDELT2": 0.5,
    "CRPIX2": 0,
    "CRVAL2": 0,
    "NAXIS2": 3,
    "CTYPE3": "Wavelength",
    "CUNIT3": "seconds",
    "CDELT3": 0.4,
    "CRPIX3": 0,
    "CRVAL3": 0,
    "NAXIS3": 2,
    "CTYPE4": "HPLN-TAN",
    "CUNIT4": "arcsec",
    "CDELT4": 0.5,
    "CRPIX4": 0,
    "CRVAL4": 0,
    "NAXIS4": 2,
}
wcs_4D = WCS(header=header_4D, naxis=4)
header_1D = {
    "CTYPE1": "Time    ",
    "CUNIT1": "seconds",
    "CDELT1": 0.4,
    "CRPIX1": 0,
    "CRVAL1": 0,
    "NAXIS1": 2,
}
wcs_1D = WCS(header=header_1D, naxis=1)
unit = utils.DN_UNIT["SJI"]
mask_cube = data >= 0
mask_4D = data_4D >= 0
uncertainty = np.sqrt(data)
uncertainty_2D = np.sqrt(data_2D)
uncertainty_1D = np.sqrt(data_1D)
uncertainty_4D = np.sqrt(data_4D)
times = Time(["2014-12-11T19:39:00.48", "2014-12-11T19:43:07.6"])
exposure_times = 2 * np.ones((2), float) * u.s
meta = Meta({"exposure time": exposure_times}, axes={"exposure time": 0}, data_shape=data.shape)
extra_coords = [("time", 0, times)]
scaled_T = True
scaled_F = False
cube = SJICube(
    data,
    wcs,
    uncertainty=uncertainty,
    mask=mask_cube,
    unit=unit,
    scaled=scaled_T,
    meta=meta,
)
cube.extra_coords.add(*extra_coords[0])
cube_2D = SJICube(
    data_2D,
    wcs_2D,
    uncertainty=uncertainty_2D,
    mask=mask_cube,
    unit=unit,
    scaled=scaled_T,
    meta=meta,
)
cube_2D.extra_coords.add(*extra_coords[0])
cube_1D = SJICube(
    data_1D,
    wcs_1D,
    uncertainty=uncertainty_1D,
    mask=mask_cube,
    unit=unit,
    scaled=scaled_T,
    meta=meta,
)
cube_1D.extra_coords.add(*extra_coords[0])
cube_F = SJICube(
    data,
    wcs,
    uncertainty=uncertainty,
    mask=mask_cube,
    unit=unit,
    scaled=scaled_F,
    meta=meta,
)
cube_F.extra_coords.add(*extra_coords[0])
cube_4D = SJICube(
    data_4D,
    wcs_4D,
    uncertainty=uncertainty_4D,
    mask=mask_4D,
    unit=unit,
    scaled=scaled_T,
    meta=meta,
)
cube_4D.extra_coords.add(*extra_coords[0])
data_dust = np.array(
    [
        [[-1, 2, -3, 4], [2, -200, 5, 3], [0, 1, 2, -300]],
        [[2, -200, 5, 1], [10, -5, 2, 2], [10, -3, 3, 0]],
    ],
)
header = {
    "CTYPE1": "HPLN-TAN",
    "CUNIT1": "arcsec",
    "CDELT1": 0.4,
    "CRPIX1": 0,
    "CRVAL1": 0,
    "NAXIS1": 4,
    "CTYPE2": "HPLT-TAN",
    "CUNIT2": "arcsec",
    "CDELT2": 0.5,
    "CRPIX2": 0,
    "CRVAL2": 0,
    "NAXIS2": 3,
    "CTYPE3": "Time    ",
    "CUNIT3": "seconds",
    "CDELT3": 0.3,
    "CRPIX3": 0,
    "CRVAL3": 0,
    "NAXIS3": 2,
}
wcs = WCS(header=header, naxis=3)
unit = utils.DN_UNIT["SJI"]
mask_dust = data_dust == -200
dust_mask_expected = np.array(
    [
        [
            [True, True, True, True],
            [True, True, True, True],
            [True, True, False, False],
        ],
        [[True, True, True, False], [True, True, True, True], [True, True, True, True]],
    ],
)
uncertainty = 1
times = Time(["2014-12-11T19:39:00.48", "2014-12-11T19:43:07.6"])
exposure_times = 2 * np.ones((2), float) * u.s
extra_coords = [("time", 0, times)]
scaled_T = True
meta = Meta(
    {"exposure time": exposure_times, "OBSID": 1},
    axes={"exposure time": 0},
    data_shape=data_dust.shape,
)
cube_dust = SJICube(
    data_dust,
    wcs,
    uncertainty=uncertainty,
    mask=mask_dust,
    unit=unit,
    scaled=scaled_T,
    meta=meta,
)
cube_dust.extra_coords.add(*extra_coords[0])


##############################################################################
# SJICube Tests
##############################################################################
@pytest.mark.parametrize(
    ("test_input", "expected"),
    [
        (cube, data / exposure_times[0]),
        (cube_2D, data_2D / exposure_times[0]),
        (cube_1D, data_1D / exposure_times[0]),
    ],
)
def test_SJICube_apply_exposure_time_correction(test_input, expected):
    np.testing.assert_array_equal(test_input.apply_exposure_time_correction().data, expected.value)


@pytest.mark.parametrize(("test_input", "expected"), [(cube, data * exposure_times[0])])
def test_SJICube_apply_exposure_time_correction_undo(test_input, expected):
    np.testing.assert_array_equal(
        test_input.apply_exposure_time_correction(undo=True, force=True).data,
        expected.value,
    )


@pytest.mark.parametrize(("test_input", "expected"), [(cube_dust, dust_mask_expected)])
def test_SJICube_apply_dust_mask(test_input, expected):
    test_input.apply_dust_mask()
    np.testing.assert_array_equal(test_input.mask, expected)
    test_input.apply_dust_mask(undo=True)
    np.testing.assert_array_equal(test_input.mask, mask_dust)



"""
==================
Mg II Dopplergrams
==================

In this example we are going to produce a Dopplergram for the Mg II k line from a IRIS 400-step raster.
The Dopplergram is obtained by subtracting the intensities at symmetrical velocity shifts from the line core (e.g. ±50 km/s).

This very large dense raster took more than three hours to complete the 400 scans (30 s exposures), which means that the spacecraft's orbital velocity changes during the observations.
This means that any precise wavelength calibration will need to correct for those shifts.
"""

import tarfile

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pooch
from astropy.coordinates import SpectralCoord
from astropy.io import fits
from scipy.constants import c
from scipy.interpolate import interp1d

from irispy.io import read_files
from irispy.utils import image_clipping

###############################################################################
# We start with getting the data. This is done by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` as to keep this example self contained
# but using your browser will also work.
#
# Using the url: http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2014/07/08/
# we are after `iris_l2_20140708_114109_3824262996_raster.tar.gz` which is ~730 MB in size.

downloaded_tar_iris_file = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2014/07/08/20140708_114109_3824262996/iris_l2_20140708_114109_3824262996_raster.tar.gz",
    known_hash=None,
)

###############################################################################
# Now to open the file using ``irispy-lmsal``.
# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float32, the data values are no longer in DN,
# but in scaled integer units that start at -2$^{16}$/2.

raster = read_files(downloaded_tar_iris_file, memmap=True, uncertainty=False)

###############################################################################
# We are after the Mg II k window, which we can select using a key.

mg_ii = raster["Mg II k 2796"][0]
(mg_wave,) = mg_ii.axis_world_coords("wl")

# We will plot the spatially averaged spectrum
plt.plot(mg_wave.to("nm"), mg_ii.data.mean((0, 1)))
plt.ylabel("DN (Memory Mapped Value)")
plt.xlabel("Wavelength (nm)")

plt.show()

###############################################################################
# To better understand the orbital velocity problem let us look at how the
# line intensity varies for a strong Mn I line at around 280.2 nm, in
# between the Mg II k and h lines. For this dataset, the line core of this
# line falls around index 350. To plot a spectrogram in the correct
# orientation we will transpose the data.

lower_corner = [SpectralCoord(280.2, unit=u.nm), None]
upper_corner = [SpectralCoord(280.2, unit=u.nm), None]
mg_crop = mg_ii.crop(lower_corner, upper_corner)

vmin, vmax = image_clipping(mg_ii.data[..., 350])
plt.figure(figsize=(6, 10))
plt.imshow(
    mg_ii.data[..., 350].T,
    origin="lower",
    aspect=0.5,
    vmin=vmin,
    vmax=vmax,
)
plt.xlabel("Solar X (arcsec)")
plt.ylabel("Solar Y (arcsec)")

plt.show()

###############################################################################
# You can see a regular bright-dark pattern along the Y axis, an
# indication that its intensities are not taken at the same position in
# the line because of wavelength shifts. The shifts are caused by the
# orbital velocity changes, and we can find these in the auxiliary
# metadata which are to be found in the extension past the "last" window
# in the FITS file.

# astropy.io.fits does not support opening tar files, so we need to extract.
tar_iris_file = tarfile.open(downloaded_tar_iris_file, "r:gz")
tar_iris_file.extractall("./")
tar_iris_file.close()
raster_filename = "iris_l2_20140708_114109_3824262996_raster_t000_r00000.fits"

# In this case, it is the 9th HDU, which we access directly
aux_data = fits.getdata(raster_filename, 9)
aux_header = fits.getheader(raster_filename, 9)
v_obs = aux_data[:, aux_header["OBS_VRIX"]]
# Convert to km/s
v_obs /= 1000.0

plt.figure()
plt.plot(v_obs)
plt.ylabel("Orbital velocity (km/s)")
plt.xlabel("Scan number")

plt.show()

###############################################################################
# To look at intensities at any given scan we only need to subtract this
# velocity shift from the wavelength scale, but to look at the whole image
# at a given wavelength we must interpolate the original data to take this
# shift into account. Here is a way to do it (note that array dimensions
# apply to this specific example).

c_kms = c / 1000.0
wave_shift = -v_obs * mg_wave[350] / (c_kms)
# Linear interpolation in wavelength, for each scan
for i in range(mg_ii.data.shape[0]):
    method = interp1d(mg_wave - wave_shift[i], mg_ii.data[:, i, :], bounds_error=False)
    mg_ii.data[:, i, :] = method(mg_wave)

###############################################################################
# Now we can plot the shifted data to see that the large scale shifts
# have disappeared

vmin, vmax = image_clipping(mg_ii.data[..., 350])
plt.figure(figsize=(6, 10))
plt.imshow(
    mg_ii.data[..., 350].T,
    origin="lower",
    aspect=0.5,
    vmin=vmin,
    vmax=vmax,
)
plt.xlabel("Solar X (arcsec)")
plt.ylabel("Solar Y (arcsec)")

plt.show()

###############################################################################
# Some residual shift remains, but we will not correct for it here. A more
# elaborate correction can be obtained by the IDL routine
# ``iris_prep_wavecorr_l2``, but this has not yet been ported to Python
# see the `IDL version of this
# tutorial <http://iris.lmsal.com/itn26/tutorials.html#mg-ii-dopplergrams>`__
# for more details.
#
# We can use the calibrated data for example to calculate Dopplergrams. A
# Dopplergram is here defined as the difference between the intensities at
# two wavelength positions at the same (and opposite) distance from the
# line core. For example, at +/- 50 km/s from the Mg II k3 core. To do
# this, let us first calculate a velocity scale for the k line and find
# the indices of the -50 and +50 km/s velocity positions (here using the
# convention of negative velocities for up flows):

mg_k_centre = 279.6351 * u.nm
pos = 50  # in km/s around line centre
velocity = (mg_wave - mg_k_centre) * c_kms / mg_k_centre
index_p = np.argmin(np.abs(velocity - pos))
index_m = np.argmin(np.abs(velocity + pos))
doppl = mg_ii.data[..., index_m] - mg_ii.data[..., index_p]

###############################################################################
# And now we can plot this as before (intensity units are again arbitrary
# because of the unscaled DNs):

vmin, vmax = image_clipping(doppl)
plt.figure(figsize=(6, 10))
plt.imshow(
    doppl.T,
    cmap="gist_gray",
    origin="lower",
    aspect=0.5,
    vmin=vmin,
    vmax=vmax,
)
plt.xlabel("Solar X (arcsec)")
plt.ylabel("Solar Y (arcsec)")

plt.show()



"""
=======================
Umbral flashes analysis
=======================

In this example, we are going to work with spectra and slit-jaw
images to study an example of dynamical phenomena called umbral flashes.
"""

import astropy.units as u
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pooch
from astropy.coordinates import SpectralCoord
from astropy.visualization import time_support

from irispy.io import read_files
from irispy.utils import image_clipping

time_support()

###############################################################################
# We start with getting the data. This is done by downloading the data from the IRIS archive.
#
# In this case, we will use ``pooch`` as to keep this example self contained
# but using your browser will also work.
#
# Using the url: http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2013/09/02/20130902_163935_4000255147/
# we are after the 1400 slit-jaw file and the raster file (~900 MB).

raster_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2013/09/02/20130902_163935_4000255147/iris_l2_20130902_163935_4000255147_raster.tar.gz",
    known_hash="e0fdbe63b6ff88856d97dff1fd4c3f62bf0bb4a85c025930c36fd69086ed0533",
)
sji_filename = pooch.retrieve(
    "http://www.lmsal.com/solarsoft/irisa/data/level2_compressed/2013/09/02/20130902_163935_4000255147/iris_l2_20130902_163935_4000255147_SJI_1400_t000.fits.gz",
    known_hash="e04eedbe7fd7e5e173b6a9308db3e45991634a4f8d9b245a334d5c492d85d0ed",
)

###############################################################################
# Now to open the files using ``irispy-lmsal``.

# Note that when ``memmap=True``, the data values are read from the FITS file
# directly without the scaling to Float32, the data values are no longer in DN,
# but in scaled integer units that start at -2$^{16}$/2.
raster = read_files(raster_filename, memmap=True, uncertainty=False)
sji_1400 = read_files(sji_filename, memmap=True, uncertainty=False)

###############################################################################
# We are after the Mg II k and C II lines, which we can select using a key.
# We will produce a space-time image of the Mg II k3 wavelength.

mg_ii = raster["Mg II k 2796"][0]
c_ii = raster["C II 1336"][0]

# Instead of using the pixel index, we can crop the data in wavelength space.
lower_corner = [SpectralCoord(279.63, unit=u.nm), None]
upper_corner = [SpectralCoord(279.63, unit=u.nm), None]
mg_crop = mg_ii.crop(lower_corner, upper_corner)

# The Mg II k3 core is located around wavelength pixel 103.
# Which we will use to clip the data.
vmin, vmax = image_clipping(mg_ii.data[..., 103])
mg_crop.plot(vmin=vmin, vmax=vmax)
plt.xlabel("Solar Y")

plt.show()

###############################################################################
# The middle section between 60"-75" is on the umbra of a sunspot, even though
# it is not obvious from this image. One can see very clearly the umbral oscillations,
# with a clear regular pattern of dark/bright streaks.
#
# Let us now load the 1400 SJI for context.

vmin, vmax = image_clipping(sji_1400[0].data)
sji_1400[0].plot(vmin=vmin, vmax=vmax)

plt.show()

###############################################################################
# The slit pixel 220 is a location on the sunspot's umbra.
# Let us plot the k3 intensity (spectral pixel 103 of ``mg_ii``) and the
# core of the brightest C II line (spectral pixel 90 of ``c_ii``) vs
# time in minutes (showing first ~10 minutes only)

mg_ii_times = mg_ii.axis_world_coords("time", wcs=mg_ii.extra_coords)[0][:200]
c_ii_times = c_ii.axis_world_coords("time", wcs=c_ii.extra_coords)[0][:200]

plt.figure()
plt.plot(mg_ii_times, mg_ii.data[:200, 220, 103], label="Mg II k3")
(ax,) = plt.plot(c_ii_times, c_ii.data[:200, 220, 90], label="C II")
plt.legend()
plt.ylabel("DN (Memory Mapped Value)")
plt.xlabel("Time (UTC)")
ax.axes.xaxis.set_major_formatter(mdates.ConciseDateFormatter(ax.axes.xaxis.get_major_locator()))
# Rotates and right-aligns the x labels so they don't crowd each other.
for label in ax.axes.get_xticklabels(which="major"):
    label.set(rotation=30, horizontalalignment="right")

plt.show()

###############################################################################
# Imagine now you wanted to compare these oscillations with
# the intensity from the SJI. The SJI images are typically
# taken at a different cadence, so you will need to load the corresponding
# time array for the 1400 SJI.

# We will take the first 50 to cut down on the size of the data for this example.
# This is memory intensive and that an upstream bug needs to be fixed.
times_sji = sji_1400.time[:50]

###############################################################################
# Now we can plot both spectral lines and SJI for a pixel close to the slit
# at the same Y position (pre-worked out to be at index 220).

plt.figure()
plt.plot(mg_ii_times, mg_ii.data[:200, 220, 103], label="Mg II k3")
plt.plot(c_ii_times, c_ii.data[:200, 220, 90], label="C II")
(ax,) = plt.plot(times_sji, sji_1400.data[:50, 190, 220], label="1400 SJI")
plt.legend()
plt.ylabel("Counts (Memory Mapped Value)")
plt.xlabel("Time (UTC)")
ax.axes.xaxis.set_major_formatter(mdates.ConciseDateFormatter(ax.axes.xaxis.get_major_locator()))
# Rotates and right-aligns the x labels so they don't crowd each other.
for label in ax.axes.get_xticklabels(which="major"):
    label.set(rotation=30, horizontalalignment="right")

plt.show()

###############################################################################
# You are now ready to explore all the correlations, anti-correlations,
# and phase differences.
